@@ -20,8 +20,8 @@ to log, configure and control the run-time of your GUI app via command line argu
     please see the documentation of :ref:`config-options` and :ref:`config-files` in the :mod:`ae.console` namespace
     portion/module for more detailed information.
 
-:class:`MainAppBase` adds on top of the :class:`~ae.console.ConsoleApp` the concepts of :ref:`application status` and
-:ref:`application flow`, explained further down.
+:class:`MainAppBase` adds on top of the :class:`~ae.console.ConsoleApp` the concepts of :ref:`application events`,
+:ref:`application status` and :ref:`application flow`, explained further down.
 
 
 application events
@@ -191,7 +191,7 @@ variables or to adapt the app environment.
 application flow
 ----------------
 
-to control the current state and application/work flow (or context) of your application, and to persist it until the
+to control the current state and UX flow (or context) of your application, and to persist it until the
 next app start, :class:`MainBaseApp` provides two :ref:`app-state-variables`: :attr:`~MainAppBase.flow_id` to store the
 currently working flow and :attr:`~MainAppBase.flow_path` to store the history of nested flows.
 
@@ -227,12 +227,13 @@ event handler:
 * `'suggest'` for autocompletion or other suggestions.
 
 all other flow actions need to confirmed to be changed by :meth:`~MainAppBase.change_flow`, either by a custom flow
-change confirmation method/event-handler or by declaring a popup class that has the same name returned by
-:func:`flow_popup_class_name` if called with the new flow id.
+change confirmation method/event-handler or by declaring a popup class. the name of the event handler and of the
+popup class gets determined from the flow id.
 
-the name of the flow change confirmation method that gets fired when the app want to change the flow (via the method
-:meth:`~MainAppBase.change_flow`) depends on the new flow id and gets determined by the function
-:func:`flow_change_confirmation_event_name`.
+.. hint::
+    the name of the flow change confirmation method that gets fired when the app want to change the flow (via the method
+    :meth:`~MainAppBase.change_flow`) gets determined by the function :func:`flow_change_confirmation_event_name`,
+    whereas the name of the popup class get determined by the function :func:`flow_popup_class_name`.
 
 if the flow-specific change confirmation event handler does not exist or returns in a boolean `False` or `None` then
 :meth:`~MainAppBase.on_flow_change` will be called. if this call also returns `False` then the action of the new flow id
@@ -250,6 +251,27 @@ finally, if a confirmed flow change results in a `'focus'` flow action then the
 fired. this event can be used by the GUI framework to set the focus to the widget associated with the new focus flow id.
 
 
+flow actions `'open'` and `'close'`
+___________________________________
+
+to display an instance of a properly named popup class, simply initiate the change the app flow to an appropriate
+flow id (with an `'open'` flow action). in this case no change confirmation event handler is needed, because
+:meth:`~MainAppBase.on_flow_change` is then automatically opening the popup.
+
+when the popup is visible the flow path will be extended with the respective flow id.
+
+calling the `close` method of the popup will hide it. on closing the popup the flow id will be reset and the opening
+flow id will be removed from the flow path.
+
+all popup classes are providing the events `on_pre_open`, `on_open`, `on_pre_dismiss` and `on_dismiss`.
+the `on_dismiss` event handler can be used for data validation: returning a non-False value from it will cancel
+the close.
+
+.. hint::
+    see the documentation of each popup class for more details on the features of popup classes (for Kivy apps e.g.
+    :class:`~ae.kivy_app.FlowDropDown`, :class:`~ae.kivy_app.FlowPopup` or :class:`~ae.kivy_app.FlowSelector`).
+
+
 key press events
 ----------------
 
@@ -407,7 +429,7 @@ from ae.core import DEBUG_LEVELS, registered_app_names
 from ae.console import ConsoleApp                                                                       # type: ignore
 
 
-__version__ = '0.3.89'
+__version__ = '0.3.90'
 
 
 APP_STATE_SECTION_NAME = 'aeAppState'           #: config section name to store app state
@@ -491,14 +513,14 @@ def ensure_tap_kwargs_refs(init_kwargs: EventKwargsType, tap_widget: Any):
         if 'popup_kwargs' not in tap_kwargs:
             tap_kwargs['popup_kwargs'] = {}
         popup_kwargs = tap_kwargs['popup_kwargs']
-        if 'parent' not in popup_kwargs:
-            popup_kwargs['parent'] = tap_kwargs['tap_widget']
+        if 'opener' not in popup_kwargs:
+            popup_kwargs['opener'] = tap_kwargs['tap_widget']
 
     """
     init_kwargs['tap_kwargs'] = tap_kwargs = init_kwargs.get('tap_kwargs', {})
     tap_kwargs['tap_widget'] = tap_widget = tap_kwargs.get('tap_widget', tap_widget)
     tap_kwargs['popup_kwargs'] = popup_kwargs = tap_kwargs.get('popup_kwargs', {})
-    popup_kwargs['parent'] = popup_kwargs.get('parent', tap_widget)
+    popup_kwargs['opener'] = popup_kwargs.get('opener', tap_widget)
 
 
 def flow_action(flow_id: str) -> str:
@@ -613,7 +635,7 @@ def flow_popup_class_name(flow_id: str) -> str:
     return flow_class_name(flow_id, 'Popup')
 
 
-def id_of_flow(action: str, obj: str = "", key: str = "") -> str:
+def id_of_flow(action: str, obj: str = '', key: str = '') -> str:
     """ create flow id string.
 
     :param action:              flow action string.
@@ -863,10 +885,9 @@ class MainAppBase(ConsoleApp, ABC):
                                   :paramref:`~change_flow.new_flow_id` argument, but then overwrite this flow id with
                                   this event arg value to set :attr:`~MainAppBase.flow_id`.
                                 * `popup_kwargs`: optional dict passed to the Popup `__init__` method, like e.g.
-                                  dict(parent=parent_widget_of_popup, data=...).
+                                  dict(opener=opener_widget_of_popup, data=...).
                                 * `popups_to_close`: optional sequence of widgets to be closed by this method after flow
-                                  change got confirmed. to prevent weakref-errors the `attach_to` attribute of each
-                                  widget will additionally be reset to None.
+                                  change got confirmed.
                                 * 'reset_last_focus_flow_id': pass `True` to reset the last focus flow id, pass `False`
                                   or `None` to ignore the last focus id (and not use to set flow id) or pass a flow id
                                   string value to change the last focus flow id to the passed value.
@@ -899,9 +920,8 @@ class MainAppBase(ConsoleApp, ABC):
             if action == 'edit' and self.flow_path_action() == 'edit' \
                     and flow_key_split(self.flow_path[-1])[0] == flow_key_split(new_flow_id)[0]:
                 _flow_id = self.flow_path.pop()
-                self.vpo(f"{prefix}PATH EDIT: removed '{_flow_id}' and adding {new_flow_id} to {self.flow_path}")
-            else:
-                self.vpo(f"{prefix}PATH EXTEND: appending '{new_flow_id}' to {self.flow_path}")
+                self.vpo(f"{prefix}PATH EDIT: removed '{_flow_id}', to replace it with ...")
+            self.vpo(f"{prefix}PATH EXTEND: appending '{new_flow_id}' to {self.flow_path}")
             self.flow_path.append(new_flow_id)
             self.change_app_state('flow_path', self.flow_path)
             flow_id = empty_flow if action == 'enter' else new_flow_id
@@ -909,7 +929,7 @@ class MainAppBase(ConsoleApp, ABC):
             # dismiss gets sent sometimes twice (e.g. on heavy double-clicking on drop-down-open-buttons)
             # .. therefore prevent run-time error
             if not self.flow_path:
-                self.vpo(f"{prefix}PATH REDUCE: cancelled flow change because of empty flow path")
+                self.dpo(f"{prefix}CANCELLED change_flow({new_flow_id}, {event_kwargs}) because of empty flow path")
                 return False
             ended_flow_id = self.flow_path.pop()
             self.vpo(f"{prefix}PATH REDUCE: popped '{ended_flow_id}' from {self.flow_path}")
@@ -924,8 +944,8 @@ class MainAppBase(ConsoleApp, ABC):
         popups_to_close = event_kwargs.get('popups_to_close', ())
         for widget in reversed(popups_to_close):
             widget.close()
-            if hasattr(widget, 'attach_to'):
-                widget.attach_to = None     # mainly for DropDown widgets to prevent weakref-error/-exception
+            # REMOVED in ae.gui_app v0.3.90: if hasattr(widget, 'attach_to'):
+            #    widget.attach_to = None     # mainly for DropDown widgets to prevent weakref-error/-exception
 
         if action not in ACTIONS_REDUCING_FLOW_PATH or not has_flow_focus:
             flow_id = event_kwargs.get('flow_id', flow_id)  # update flow_id from event_kwargs
@@ -965,9 +985,16 @@ class MainAppBase(ConsoleApp, ABC):
         """ dpi scaling factor - override if the used GUI framework supports dpi scaling. """
         return 1.0
 
-    def close_popups(self):
-        """ close all opened popups (starting with the topmost popup). """
-        for popup in self.popups_opened():
+    def close_popups(self, classes: tuple = ()):
+        """ close all opened popups (starting with the foremost popup).
+
+        :param classes:         optional class filter - if not passed then only the first foremost widgets underneath
+                                the app win with an `open` method will be closed. pass tuple to restrict found popup
+                                widgets to certain classes. like e.g. by passing (Popup, DropDown, FlowPopup) to get
+                                all popups of an app (in Kivy use Factory.WidgetClass if widget is declared only in
+                                kv lang).
+        """
+        for popup in self.popups_opened(classes=classes):
             popup.close()
 
     def find_image(self, image_name: str, height: float = 32.0, light_theme: bool = True) -> Optional[RegisteredFile]:
@@ -1009,14 +1036,14 @@ class MainAppBase(ConsoleApp, ABC):
         return None
 
     def find_widget(self, match: Callable[[Any], bool]) -> Optional[Any]:
-        """ search the widget tree returning the first matching widget in reversed z-order (top-most first).
+        """ search the widget tree returning the first matching widget in reversed z-order (top-/foremost first).
 
         :param match:           callable called with the widget as argument, returning True if widget matches.
-        :return:                first found widget in reversed z-order (top-most widget first).
+        :return:                first found widget in reversed z-order (top-/fore-most widget first).
         """
         def child_wid(children):
             """ bottom up search within children for a widget with matching attribute name and value. """
-            for widget in reversed(children):
+            for widget in children:
                 found = child_wid(self.widget_children(widget))
                 if found:
                     return found
@@ -1351,17 +1378,19 @@ class MainAppBase(ConsoleApp, ABC):
         :param classes:         optional class filter - if not passed then only the widgets underneath win/root with an
                                 `open` method will be added. pass tuple to restrict found popup widgets to certain
                                 classes. like e.g. by passing (Popup, DropDown, FlowPopup) to get all popups of an
-                                ae/Kivy app.
-        :return:                list of opened/visible popup class instances that are children of either the
-                                root layout or the app window, ordered by their z-coordinate (most front widget last).
+                                ae/Kivy app (in Kivy use Factory.WidgetClass if widget is declared only in kv lang).
+        :return:                list of the foremost opened/visible popup class instances (children of the app window),
+                                matching the :paramref:`classes` or having an `open` method, ordered by their
+                                z-coordinate (most front widget first).
         """
         filter_func = (lambda _wg: isinstance(_wg, classes)) if classes else \
             (lambda _wg: callable(getattr(_wg, 'open', None)))
 
         popups = []
-        for wid in self.framework_win.children + self.framework_root.children:
-            if filter_func(wid):
-                popups.append(wid)
+        for wid in self.framework_win.children:  # REMOVED in ae.gui_app v0.3.90: + self.framework_root.children:
+            if not filter_func(wid):
+                break
+            popups.append(wid)
 
         return popups
 
@@ -1531,7 +1560,7 @@ class MainAppBase(ConsoleApp, ABC):
         return self.widget_by_attribute('app_state_name', app_state_name)
 
     def widget_children(self, wid: Any, only_visible: bool = False) -> List:
-        """ determine the children of widget or its container (if exists) in z-order (top-most last).
+        """ determine the children of widget or its container (if exists) in z-order (top-/foremost first).
 
         :param wid:             widget to determine the children from.
         :param only_visible:    pass True to only return visible widgets.
